'''
Packaging
=========

This module contains `PyInstaller <http://www.pyinstaller.org/>`_ hooks in
order to assist in the process of building binary packages. PyInstaller allows
you to produce stand-alone, self-contained executables of your Kivy app for
Windows, Linux and Mac.

For more information, please see the `PyInstaller website
<http://www.pyinstaller.org/>`_
'''
